import { Vec2 } from "./Vec2.js";
import { once } from "../Util.js";
export class Range2 {
    get isZero() {
        return this.pos.isZero && this.size.isZero;
    }
    constructor(x = 0, y = 0, w = 0, h = 0) {
        this.pos = new Vec2(x, y);
        this.size = new Vec2(w, h);
    }
    create(x = 0, y = 0, w = 0, h = 0) {
        return new (this.constructor)(x, y, w, h);
    }
    set(src) {
        once(() => {
            console.warn("Function Range2::set is deprecated please use Range2::copy instead. get is a reserved word.");
        });
        return this.copy(src);
    }
    copy(src) {
        this.pos.copy(src.pos);
        this.size.copy(src.size);
        return this;
    }
    clone(out) {
        const result = out ? out.copy(this) : this.create(this.pos.x, this.pos.y, this.size.x, this.size.y);
        return result;
    }
    fromRect(rect) {
        let start = rect.start;
        let stop = rect.stop;
        this.pos.x = start.x;
        this.pos.y = start.y;
        this.size.x = stop.x + (rect.endInclusive ? (stop.x < start.x ? -1 : 1) : 0) - start.x;
        this.size.y = stop.y + (rect.endInclusive ? (stop.y < start.y ? -1 : 1) : 0) - start.y;
        return this;
    }
    scale(factor, keepCenter = true) {
        let org = null;
        if (keepCenter) {
            org = this.size.clone();
        }
        this.size.scale(factor);
        if (keepCenter) {
            this.pos.add(org.subtract(this.size).multiply(0.5));
        }
        return this;
    }
    translate(system) {
        this.pos.scale(system);
        this.size.scale(system);
        return this;
    }
    move(system) {
        this.pos.add(system);
        return this;
    }
    toInt() {
        this.pos.toInt();
        this.size.toInt();
        return this;
    }
    toDecimal() {
        this.pos.toDecimal();
        this.size.toDecimal();
        return this;
    }
    contains(target) {
        return this.pos.x <= target.pos.x &&
            this.pos.y <= target.pos.y &&
            this.pos.x + this.size.x >= target.pos.x + target.size.x &&
            this.pos.y + this.size.y >= target.pos.y + target.size.y;
    }
    containsPoint(vec) {
        return vec.x >= this.pos.x && vec.x <= this.pos.x + this.size.x - 1
            && vec.y >= this.pos.y && vec.y <= this.pos.y + this.size.y - 1;
    }
    first(fn) {
        const p = new Vec2();
        const x = this.pos.x;
        const y = this.pos.y;
        for (let i = 0; i < this.size.x; i++) {
            for (let j = 0; j < this.size.y; j++) {
                p.x = i + x, p.y = j + y;
                if (fn(p)) {
                    return p;
                }
            }
        }
        return null;
    }
    forEach(fn, start = null) {
        const pos = new Vec2();
        const begin = this.pos.clone().toInt();
        if (start === null || !this.containsPoint(start)) {
            start = begin;
        }
        const end = this.pos.clone().add(this.size).toInt();
        for (let y = begin.y; y < end.y; y += 1) {
            for (let x = begin.x; x < end.x; x += 1) {
                if (y < start.y || (y === start.y && x < start.x)) {
                    continue;
                }
                pos.x = x;
                pos.y = y;
                const brk = fn(pos);
                if (brk) {
                    return;
                }
            }
        }
    }
    equals(range) {
        return this.pos.equals(range.pos) && this.size.equals(range.size);
    }
    zero() {
        this.pos.zero();
        this.size.zero();
        return this;
    }
}
//# sourceMappingURL=Range2.js.map